# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""Common utilities for conda command line tools."""
import re
import sys
from logging import getLogger
from os.path import basename, dirname, isdir, isfile, join, normcase

from ..auxlib.ish import dals
from ..base.constants import ROOT_ENV_NAME
from ..base.context import context, env_name
from ..common.constants import NULL
from ..common.io import swallow_broken_pipe
from ..common.path import paths_equal
from ..common.serialize import json_dump
from ..exceptions import (
    CondaError,
    DirectoryNotACondaEnvironmentError,
    EnvironmentLocationNotFound,
)
from ..models.match_spec import MatchSpec


def confirm(message="Proceed", choices=("yes", "no"), default="yes", dry_run=NULL):
    assert default in choices, default
    if (dry_run is NULL and context.dry_run) or dry_run:
        from ..exceptions import DryRunExit

        raise DryRunExit()

    options = []
    for option in choices:
        if option == default:
            options.append("[%s]" % option[0])
        else:
            options.append(option[0])
    message = "{} ({})? ".format(message, "/".join(options))
    choices = {alt: choice for choice in choices for alt in [choice, choice[0]]}
    choices[""] = default
    while True:
        # raw_input has a bug and prints to stderr, not desirable
        sys.stdout.write(message)
        sys.stdout.flush()
        try:
            user_choice = sys.stdin.readline().strip().lower()
        except OSError as e:
            raise CondaError(f"cannot read from stdin: {e}")
        if user_choice not in choices:
            print("Invalid choice: %s" % user_choice)
        else:
            sys.stdout.write("\n")
            sys.stdout.flush()
            return choices[user_choice]


def confirm_yn(message="Proceed", default="yes", dry_run=NULL):
    if (dry_run is NULL and context.dry_run) or dry_run:
        from ..exceptions import DryRunExit

        raise DryRunExit()
    if context.always_yes:
        return True
    try:
        choice = confirm(
            message=message, choices=("yes", "no"), default=default, dry_run=dry_run
        )
    except KeyboardInterrupt:  # pragma: no cover
        from ..exceptions import CondaSystemExit

        raise CondaSystemExit("\nOperation aborted.  Exiting.")
    if choice == "no":
        from ..exceptions import CondaSystemExit

        raise CondaSystemExit("Exiting.")
    return True


def is_active_prefix(prefix: str) -> bool:
    """
    Determines whether the args we pass in are pointing to the active prefix.
    Can be used a validation step to make sure operations are not being
    performed on the active prefix.
    """
    if context.active_prefix is None:
        return False
    return (
        paths_equal(prefix, context.active_prefix)
        # normcasing our prefix check for Windows, for case insensitivity
        or normcase(prefix) == normcase(env_name(context.active_prefix))
    )


def arg2spec(arg, json=False, update=False):
    try:
        spec = MatchSpec(arg)
    except:
        from ..exceptions import CondaValueError

        raise CondaValueError("invalid package specification: %s" % arg)

    name = spec.name
    if not spec._is_simple() and update:
        from ..exceptions import CondaValueError

        raise CondaValueError(
            """version specifications not allowed with 'update'; use
    conda update  {}{}  or
    conda install {}""".format(
                name, " " * (len(arg) - len(name)), arg
            )
        )

    return str(spec)


def specs_from_args(args, json=False):
    return [arg2spec(arg, json=json) for arg in args]


spec_pat = re.compile(
    r"""
    (?P<name>[^=<>!\s]+)                # package name
    \s*                                 # ignore spaces
    (
        (?P<cc>=[^=]+(=[^=]+)?)         # conda constraint
        |
        (?P<pc>(?:[=!]=|[><]=?|~=).+)   # new pip-style constraints
    )?$
    """,
    re.VERBOSE,
)


def strip_comment(line):
    return line.split("#")[0].rstrip()


def spec_from_line(line):
    m = spec_pat.match(strip_comment(line))
    if m is None:
        return None
    name, cc, pc = (m.group("name").lower(), m.group("cc"), m.group("pc"))
    if cc:
        return name + cc.replace("=", " ")
    elif pc:
        if pc.startswith("~= "):
            assert (
                pc.count("~=") == 1
            ), f"Overly complex 'Compatible release' spec not handled {line}"
            assert pc.count("."), f"No '.' in 'Compatible release' version {line}"
            ver = pc.replace("~= ", "")
            ver2 = ".".join(ver.split(".")[:-1]) + ".*"
            return name + " >=" + ver + ",==" + ver2
        else:
            return name + " " + pc.replace(" ", "")
    else:
        return name


def specs_from_url(url, json=False):
    from ..gateways.connection.download import TmpDownload

    explicit = False
    with TmpDownload(url, verbose=False) as path:
        specs = []
        try:
            for line in open(path):
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                if line == "@EXPLICIT":
                    explicit = True
                if explicit:
                    specs.append(line)
                    continue
                spec = spec_from_line(line)
                if spec is None:
                    from ..exceptions import CondaValueError

                    raise CondaValueError(f"could not parse '{line}' in: {url}")
                specs.append(spec)
        except OSError as e:
            from ..exceptions import CondaFileIOError

            raise CondaFileIOError(path, e)
    return specs


def names_in_specs(names, specs):
    return any(spec.split()[0] in names for spec in specs)


def disp_features(features):
    if features:
        return "[%s]" % " ".join(features)
    else:
        return ""


@swallow_broken_pipe
def stdout_json(d):
    getLogger("conda.stdout").info(json_dump(d))


def stdout_json_success(success=True, **kwargs):
    result = {"success": success}
    actions = kwargs.pop("actions", None)
    if actions:
        if "LINK" in actions:
            actions["LINK"] = [prec.dist_fields_dump() for prec in actions["LINK"]]
        if "UNLINK" in actions:
            actions["UNLINK"] = [prec.dist_fields_dump() for prec in actions["UNLINK"]]
        result["actions"] = actions
    result.update(kwargs)
    stdout_json(result)


def print_envs_list(known_conda_prefixes, output=True):
    if output:
        print("# conda environments:")
        print("#")

    def disp_env(prefix):
        fmt = "%-20s  %s  %s"
        active = "*" if prefix == context.active_prefix else " "
        if prefix == context.root_prefix:
            name = ROOT_ENV_NAME
        elif any(
            paths_equal(envs_dir, dirname(prefix)) for envs_dir in context.envs_dirs
        ):
            name = basename(prefix)
        else:
            name = ""
        if output:
            print(fmt % (name, active, prefix))

    for prefix in known_conda_prefixes:
        disp_env(prefix)

    if output:
        print()


def check_non_admin():
    from ..common._os import is_admin

    if not context.non_admin_enabled and not is_admin():
        from ..exceptions import OperationNotAllowed

        raise OperationNotAllowed(
            dals(
                """
            The create, install, update, and remove operations have been disabled
            on your system for non-privileged users.
        """
            )
        )


def validate_prefix(prefix):
    """Verifies the prefix is a valid conda environment.

    :raises EnvironmentLocationNotFound: Non-existent path or not a directory.
    :raises DirectoryNotACondaEnvironmentError: Directory is not a conda environment.
    :returns: Valid prefix.
    :rtype: str
    """
    if isdir(prefix):
        if not isfile(join(prefix, "conda-meta", "history")):
            raise DirectoryNotACondaEnvironmentError(prefix)
    else:
        raise EnvironmentLocationNotFound(prefix)

    return prefix



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""Common utilities for conda-env command line tools."""
from os.path import abspath, expanduser, expandvars

from conda.base.context import context
from conda.cli import install as cli_install
from conda.cli.common import stdout_json_success
from conda.gateways.connection.session import CONDA_SESSION_SCHEMES

base_env_name = "base"


def print_result(args, prefix, result):
    if context.json:
        if result["conda"] is None and result["pip"] is None:
            stdout_json_success(message="All requested packages already installed.")
        else:
            if result["conda"] is not None:
                actions = result["conda"]
            else:
                actions = {}
            if result["pip"] is not None:
                actions["PIP"] = result["pip"]
            stdout_json_success(prefix=prefix, actions=actions)
    else:
        cli_install.print_activate(args.name or prefix)


def get_filename(filename):
    """Expand filename if local path or return the url"""
    url_scheme = filename.split("://", 1)[0]
    if url_scheme in CONDA_SESSION_SCHEMES:
        return filename
    else:
        return abspath(expanduser(expandvars(filename)))



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
import os
import random
from io import StringIO
from pathlib import Path
from unittest.mock import patch
from uuid import uuid4

import pytest
from pytest import MonkeyPatch

from conda.base.context import context, reset_context
from conda.common.serialize import yaml_round_trip_load
from conda.exceptions import CondaHTTPError, EnvironmentFileNotFound
from conda.models.match_spec import MatchSpec
from conda.testing import CondaCLIFixture
from conda.testing.integration import package_is_installed
from conda_env.env import (
    VALID_KEYS,
    Environment,
    from_environment,
    from_file,
    load_from_directory,
)

from . import support_file
from .utils import make_temp_envs_dir


class FakeStream:
    def __init__(self):
        self.output = ""

    def write(self, chunk):
        self.output += chunk.decode("utf-8")


def get_environment(filename):
    return from_file(support_file(filename))


def get_simple_environment():
    return get_environment("simple.yml")


def get_valid_keys_environment():
    return get_environment("valid_keys.yml")


def get_invalid_keys_environment():
    return get_environment("invalid_keys.yml")


def test_returns_Environment():
    e = get_simple_environment()
    assert isinstance(e, Environment)


def test_retains_full_filename():
    e = get_simple_environment()
    assert support_file("simple.yml") == e.filename


def test_with_pip():
    e = from_file(support_file("with-pip.yml"))
    assert "pip" in e.dependencies
    assert "foo" in e.dependencies["pip"]
    assert "baz" in e.dependencies["pip"]


@pytest.mark.timeout(20)
def test_add_pip():
    e = from_file(support_file("add-pip.yml"))
    expected = {
        "conda": ["pip", "car"],
        "pip": ["foo", "baz"],
    }
    assert e.dependencies == expected


@pytest.mark.integration
def test_http():
    e = get_simple_environment()
    f = from_file(
        "https://raw.githubusercontent.com/conda/conda/main/tests/conda_env/support/simple.yml"
    )
    assert e.dependencies == f.dependencies
    assert e.dependencies == f.dependencies


@pytest.mark.integration
def test_http_raises():
    with pytest.raises(CondaHTTPError):
        from_file(
            "https://raw.githubusercontent.com/conda/conda/main/tests/conda_env/support/does-not-exist.yml"
        )


def test_envvars():
    current_conda_token = os.environ.get("CONDA_TOKEN")
    os.environ["CONDA_TOKEN"] = "aaa-12345"
    os.environ["OTHER_KEY"] = "12345-aaa"
    e = get_environment("channels_with_envvars.yml")
    assert set(e.channels) == {
        "https://localhost/t/aaa-12345/stable",
        "https://localhost/t/12345-aaa/stable",
        "conda-forge",
        "defaults",
    }
    if current_conda_token:
        os.environ["CONDA_TOKEN"] = current_conda_token
    else:
        del os.environ["CONDA_TOKEN"]
    del os.environ["OTHER_KEY"]


def test_has_empty_filename_by_default():
    e = Environment()
    assert e.filename is None


def test_has_filename_if_provided():
    r = random.randint(100, 200)
    random_filename = f"/path/to/random/environment-{r}.yml"
    e = Environment(filename=random_filename)
    assert e.filename == random_filename


def test_has_empty_name_by_default():
    e = Environment()
    assert e.name is None


def test_has_name_if_provided():
    random_name = f"random-{random.randint(100, 200)}"
    e = Environment(name=random_name)
    assert e.name == random_name


def test_dependencies_are_empty_by_default():
    e = Environment()
    assert not e.dependencies


def test_parses_dependencies_from_raw_file():
    e = get_simple_environment()
    expected = {"conda": ["nltk"]}
    assert e.dependencies == expected


def test_builds_spec_from_line_raw_dependency():
    # TODO Refactor this inside conda to not be a raw string
    e = Environment(dependencies=["nltk=3.0.0=np18py27_0"])
    expected = {"conda": ["nltk==3.0.0=np18py27_0"]}
    assert e.dependencies == expected


def test_args_are_wildcarded():
    e = Environment(dependencies=["python=2.7"])
    expected = {"conda": ["python=2.7"]}
    assert e.dependencies == expected


def test_other_tips_of_dependencies_are_supported():
    e = Environment(dependencies=["nltk", {"pip": ["foo", "bar"]}])
    expected = {
        "conda": ["nltk", "pip"],
        "pip": ["foo", "bar"],
    }
    assert e.dependencies == expected


def test_channels_default_to_empty_list():
    e = Environment()
    assert isinstance(e.channels, list)
    assert not e.channels


def test_add_channels():
    e = Environment()
    e.add_channels(["dup", "dup", "unique"])
    assert e.channels == ["dup", "unique"]


def test_remove_channels():
    e = Environment(channels=["channel"])
    e.remove_channels()
    assert not e.channels


def test_channels_are_provided_by_kwarg():
    random_channels = (random.randint(100, 200), random)
    e = Environment(channels=random_channels)
    assert e.channels == random_channels


def test_to_dict_returns_dictionary_of_data():
    random_name = f"random{random.randint(100, 200)}"
    e = Environment(name=random_name, channels=["javascript"], dependencies=["nodejs"])

    expected = {
        "name": random_name,
        "channels": ["javascript"],
        "dependencies": ["nodejs"],
    }
    assert e.to_dict() == expected


def test_to_dict_returns_just_name_if_only_thing_present():
    e = Environment(name="simple")
    expected = {"name": "simple"}
    assert e.to_dict() == expected


def test_to_yaml_returns_yaml_parseable_string():
    random_name = f"random{random.randint(100, 200)}"
    e = Environment(name=random_name, channels=["javascript"], dependencies=["nodejs"])

    expected = {
        "name": random_name,
        "channels": ["javascript"],
        "dependencies": ["nodejs"],
    }

    actual = yaml_round_trip_load(StringIO(e.to_yaml()))
    assert expected == actual


def test_to_yaml_returns_proper_yaml():
    random_name = f"random{random.randint(100, 200)}"
    e = Environment(name=random_name, channels=["javascript"], dependencies=["nodejs"])

    expected = "\n".join(
        [
            "name: %s" % random_name,
            "channels:",
            "  - javascript",
            "dependencies:",
            "  - nodejs",
            "",
        ]
    )

    actual = e.to_yaml()
    assert expected == actual


def test_to_yaml_takes_stream():
    random_name = f"random{random.randint(100, 200)}"
    e = Environment(name=random_name, channels=["javascript"], dependencies=["nodejs"])

    s = FakeStream()
    e.to_yaml(stream=s)

    expected = "\n".join(
        [
            "name: %s" % random_name,
            "channels:",
            "  - javascript",
            "dependencies:",
            "  - nodejs",
            "",
        ]
    )
    assert expected == s.output


def test_can_add_dependencies_to_environment():
    e = get_simple_environment()
    e.dependencies.add("bar")

    s = FakeStream()
    e.to_yaml(stream=s)

    expected = "\n".join(["name: nlp", "dependencies:", "  - nltk", "  - bar", ""])
    assert expected == s.output


def test_dependencies_update_after_adding():
    e = get_simple_environment()
    assert "bar" not in e.dependencies["conda"]
    e.dependencies.add("bar")
    assert "bar" in e.dependencies["conda"]


def test_valid_keys():
    e = get_valid_keys_environment()
    e_dict = e.to_dict()
    for key in VALID_KEYS:
        assert key in e_dict


def test_invalid_keys():
    e = get_invalid_keys_environment()
    e_dict = e.to_dict()
    assert "name" in e_dict
    assert len(e_dict) == 1


@pytest.mark.parametrize(
    "directory",
    ["example", "example-yaml", "foo/bar", "foo/bar/baz", "foo/bar/baz/"],
)
def test_load_from_directory(directory: str):
    env = load_from_directory(support_file(directory))

    assert isinstance(env, Environment)
    assert "test" == env.name
    assert len(env.dependencies["conda"]) == 1
    assert "numpy" in env.dependencies["conda"]


def test_raises_when_unable_to_find():
    with pytest.raises(EnvironmentFileNotFound):
        load_from_directory("/path/to/unknown/env-spec")


def test_raised_exception_has_environment_yml_as_file():
    with pytest.raises(EnvironmentFileNotFound) as err:
        load_from_directory("/path/to/unknown/env-spec")
    assert err.value.filename == "environment.yml"


def test_load_from_directory_and_save(tmp_path: Path):
    original = Path(support_file("saved-env/environment.yml")).read_text()

    tmp = tmp_path / "environment.yml"
    tmp.write_text(original)

    env = load_from_directory(tmp)

    assert len(env.dependencies["conda"]) == 1
    assert "numpy" not in env.dependencies["conda"]

    env.dependencies.add("numpy")
    env.save()

    new_env = load_from_directory(tmp)
    assert len(new_env.dependencies["conda"]) == 2
    assert "numpy" in new_env.dependencies["conda"]


def test_creates_file_on_save(tmp_path: Path):
    tmp = tmp_path / "environment.yml"

    assert not tmp.exists()

    env = Environment(filename=tmp, name="simple")
    env.save()

    assert tmp.exists()
    assert env.to_yaml() == tmp.read_text()


@pytest.mark.integration
def test_create_advanced_pip(monkeypatch: MonkeyPatch, conda_cli: CondaCLIFixture):
    monkeypatch.setenv("CONDA_DLL_SEARCH_MODIFICATION_ENABLE", "true")

    with make_temp_envs_dir() as envs_dir:
        monkeypatch.setenv("CONDA_ENVS_DIRS", envs_dir)
        reset_context()
        assert context.envs_dirs[0] == envs_dir

        env_name = str(uuid4())[:8]
        prefix = Path(envs_dir, env_name)

        conda_cli(
            *("env", "create"),
            *("--name", env_name),
            *("--file", support_file("pip_argh.yml")),
        )
        assert prefix.exists()
        assert package_is_installed(prefix, "argh==0.26.2")


def test_from_history():
    # We're not testing that get_requested_specs_map() actually works
    # assume it gives us back a dict of MatchSpecs
    with patch("conda.history.History.get_requested_specs_map") as m:
        m.return_value = {
            "python": MatchSpec("python=3"),
            "pytest": MatchSpec("pytest!=3.7.3"),
            "mock": MatchSpec("mock"),
            "yaml": MatchSpec("yaml>=0.1"),
        }
        out = from_environment("mock_env", "mock_prefix", from_history=True)
        assert "yaml[version='>=0.1']" in out.to_dict()["dependencies"]
        assert "pytest!=3.7.3" in out.to_dict()["dependencies"]
        assert len(out.to_dict()["dependencies"]) == 4

        m.assert_called()



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
import json
import os
import tempfile
from pathlib import Path

import pytest

from conda.base.constants import ROOT_ENV_NAME
from conda.base.context import context
from conda.common.serialize import yaml_safe_load
from conda.core.envs_manager import list_all_known_prefixes
from conda.exceptions import (
    CondaEnvException,
    EnvironmentFileExtensionNotValid,
    EnvironmentFileNotFound,
    EnvironmentLocationNotFound,
    SpecNotFound,
)
from conda.gateways.disk.delete import rm_rf
from conda.testing import CondaCLIFixture, PathFactoryFixture

# Environment names we use during our tests
TEST_ENV_NAME_1 = "env-1"
TEST_ENV_NAME_2 = "snowflakes"
TEST_ENV_NAME_42 = "env-42"
TEST_ENV_NAME_PIP = "env-pip"

# Environment config files we use for out tests
ENVIRONMENT_1 = f"""
name: {TEST_ENV_NAME_1}
dependencies:
  - python
channels:
  - defaults
"""

ENVIRONMENT_1_WITH_VARIABLES = f"""
name: {TEST_ENV_NAME_1}
dependencies:
  - python
channels:
  - defaults
variables:
  DUDE: woah
  SWEET: yaaa
  API_KEY: AaBbCcDd===EeFf

"""

ENVIRONMENT_2 = f"""
name: {TEST_ENV_NAME_1}
dependencies:
  - python
  - flask
channels:
  - defaults
"""

ENVIRONMENT_3_INVALID = f"""
name: {TEST_ENV_NAME_1}
dependecies:
  - python
  - flask
channels:
  - defaults
foo: bar
"""

ENVIRONMENT_PYTHON_PIP_CLICK = f"""
name: {TEST_ENV_NAME_1}
dependencies:
  - python=3
  - pip
  - pip:
    - click
channels:
  - defaults
"""

ENVIRONMENT_PYTHON_PIP_CLICK_ATTRS = f"""
name: {TEST_ENV_NAME_1}
dependencies:
  - python=3
  - pip
  - pip:
    - click
    - attrs
channels:
  - defaults
"""

ENVIRONMENT_PYTHON_PIP_NONEXISTING = f"""
name: {TEST_ENV_NAME_1}
dependencies:
  - python=3
  - pip
  - pip:
    - nonexisting_
channels:
  - defaults
"""


def escape_for_winpath(p):
    if p:
        return p.replace("\\", "\\\\")


def create_env(content, filename="environment.yml"):
    path = Path(filename)
    path.write_text(content)


def remove_env_file(filename="environment.yml"):
    Path(filename).unlink()


@pytest.fixture
def env_name_1(conda_cli: CondaCLIFixture) -> None:
    rm_rf("environment.yml")
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_1, "--yes")
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_42, "--yes")
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_PIP, "--yes")
    for env_nb in range(1, 6):
        conda_cli("env", "remove", "--name", f"envjson-{env_nb}", "--yes")

    yield

    rm_rf("environment.yml")
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_1, "--yes")
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_42, "--yes")
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_PIP, "--yes")
    for env_nb in range(1, 6):
        conda_cli("env", "remove", "--name", f"envjson-{env_nb}", "--yes")


@pytest.mark.integration
def test_conda_env_create_no_file(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Test `conda env create` without an environment.yml file
    Should fail
    """
    with pytest.raises(EnvironmentFileNotFound):
        conda_cli("env", "create")


@pytest.mark.integration
def test_conda_env_create_no_existent_file(
    env_name_1: None, conda_cli: CondaCLIFixture
):
    """
    Test `conda env create --file=not_a_file.txt` with a file that does not
    exist.
    """
    with pytest.raises(EnvironmentFileNotFound):
        conda_cli("env", "create", "--file", "not_a_file.txt")


@pytest.mark.integration
def test_conda_env_create_no_existent_file_with_name(
    env_name_1: None, conda_cli: CondaCLIFixture
):
    """
    Test `conda env create --file=not_a_file.txt` with a file that does not
    exist.
    """
    with pytest.raises(EnvironmentFileNotFound):
        conda_cli("env", "create", "--file", "not_a_file.txt", "--name", "foo")


@pytest.mark.integration
def test_create_valid_remote_env(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Test retrieving an environment using the BinstarSpec (i.e. it retrieves it from anaconda.org)

    This tests the `remote_origin` command line argument.
    """
    conda_cli("env", "create", "conda-test/env-42")
    assert env_is_created(TEST_ENV_NAME_42)

    stdout, _, _ = conda_cli("info", "--json")

    parsed = json.loads(stdout)
    assert [env for env in parsed["envs"] if env.endswith(TEST_ENV_NAME_42)]


@pytest.mark.integration
def test_create_valid_env(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Creates an environment.yml file and
    creates and environment with it
    """
    create_env(ENVIRONMENT_1)
    conda_cli("env", "create")
    assert env_is_created(TEST_ENV_NAME_1)

    stdout, _, _ = conda_cli("info", "--json")
    parsed = json.loads(stdout)
    assert [env for env in parsed["envs"] if env.endswith(TEST_ENV_NAME_1)]


@pytest.mark.integration
def test_create_dry_run_yaml(env_name_1: None, conda_cli: CondaCLIFixture):
    create_env(ENVIRONMENT_1)
    stdout, _, _ = conda_cli("env", "create", "--dry-run")
    assert not env_is_created(TEST_ENV_NAME_1)

    # Find line where the YAML output starts (stdout might change if plugins involved)
    lines = stdout.splitlines()
    for lineno, line in enumerate(lines):
        if line.startswith("name:"):
            break
    else:
        pytest.fail("Didn't find YAML data in output")

    output = yaml_safe_load("\n".join(lines[lineno:]))
    assert output["name"] == "env-1"
    assert len(output["dependencies"]) > 0


@pytest.mark.integration
def test_create_dry_run_json(env_name_1: None, conda_cli: CondaCLIFixture):
    create_env(ENVIRONMENT_1)
    stdout, _, _ = conda_cli("env", "create", "--dry-run", "--json")
    assert not env_is_created(TEST_ENV_NAME_1)

    output = json.loads(stdout)
    assert output.get("name") == "env-1"
    assert len(output["dependencies"])


@pytest.mark.integration
def test_create_valid_env_with_variables(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Creates an environment.yml file and
    creates and environment with it
    """
    create_env(ENVIRONMENT_1_WITH_VARIABLES)
    conda_cli("env", "create")
    assert env_is_created(TEST_ENV_NAME_1)

    stdout, _, _ = conda_cli(
        *("env", "config", "vars", "list"),
        *("--name", TEST_ENV_NAME_1),
        "--json",
    )
    output_env_vars = json.loads(stdout)
    assert output_env_vars == {
        "DUDE": "woah",
        "SWEET": "yaaa",
        "API_KEY": "AaBbCcDd===EeFf",
    }

    stdout, _, _ = conda_cli("info", "--json")
    parsed = json.loads(stdout)
    assert [env for env in parsed["envs"] if env.endswith(TEST_ENV_NAME_1)]


@pytest.mark.integration
def test_conda_env_create_empty_file(env_name_1: None, conda_cli: CondaCLIFixture):
    """Test `conda env create --file=file_name.yml` where file_name.yml is empty."""
    tmp_file = tempfile.NamedTemporaryFile(suffix=".yml", delete=False)

    with pytest.raises(SpecNotFound):
        conda_cli("env", "create", "--file", tmp_file.name)

    tmp_file.close()
    os.unlink(tmp_file.name)


@pytest.mark.integration
def test_conda_env_create_http(env_name_1: None, conda_cli: CondaCLIFixture):
    """Test `conda env create --file=https://some-website.com/environment.yml`."""
    conda_cli(
        *("env", "create"),
        *(
            "--file",
            "https://raw.githubusercontent.com/conda/conda/main/tests/conda_env/support/simple.yml",
        ),
    )
    try:
        assert env_is_created("nlp")
    finally:
        conda_cli("env", "remove", "--name", "nlp", "--yes")


@pytest.mark.integration
def test_update(env_name_1: None, conda_cli: CondaCLIFixture):
    create_env(ENVIRONMENT_1)
    conda_cli("env", "create")
    create_env(ENVIRONMENT_2)

    conda_cli("env", "update", "--name", TEST_ENV_NAME_1)

    stdout, _, _ = conda_cli("list", "--name", TEST_ENV_NAME_1, "flask", "--json")
    parsed = json.loads(stdout)
    assert parsed


@pytest.mark.integration
def test_name(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    # smoke test for gh-254
    Test that --name can overide the `name` key inside an environment.yml
    """
    create_env(ENVIRONMENT_1)
    env_name = "smoke-gh-254"

    # It might be the case that you need to run this test more than once!
    try:
        conda_cli("env", "remove", "--name", env_name, "--yes")
    except:
        pass

    conda_cli("env", "create", "--file", "environment.yml", "--name", env_name, "--yes")

    stdout, _, _ = conda_cli("info", "--json")

    parsed = json.loads(stdout)
    assert [env for env in parsed["envs"] if env.endswith(env_name)]


@pytest.mark.integration
def test_create_valid_env_json_output(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Creates an environment from an environment.yml file with conda packages (no pip)
    Check the json output
    """
    create_env(ENVIRONMENT_1)
    stdout, _, _ = conda_cli(
        "env", "create", "--name", "envjson-1", "--quiet", "--json", "--yes"
    )
    output = json.loads(stdout)
    assert output["success"] is True
    assert len(output["actions"]["LINK"]) > 0
    assert "PIP" not in output["actions"]


@pytest.mark.integration
def test_create_valid_env_with_conda_and_pip_json_output(
    env_name_1: None, conda_cli: CondaCLIFixture
):
    """
    Creates an environment from an environment.yml file with conda and pip dependencies
    Check the json output
    """
    create_env(ENVIRONMENT_PYTHON_PIP_CLICK)
    stdout, _, _ = conda_cli(
        "env", "create", "--name", "envjson-2", "--quiet", "--json", "--yes"
    )
    output = json.loads(stdout)
    assert len(output["actions"]["LINK"]) > 0
    assert output["actions"]["PIP"][0].startswith("click")


@pytest.mark.integration
def test_update_env_json_output(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Update an environment by adding a conda package
    Check the json output
    """
    create_env(ENVIRONMENT_1)
    conda_cli("env", "create", "--name", "envjson-3", "--json", "--yes")
    create_env(ENVIRONMENT_2)
    stdout, _, _ = conda_cli(
        "env", "update", "--name", "envjson-3", "--quiet", "--json"
    )
    output = json.loads(stdout)
    assert output["success"] is True
    assert len(output["actions"]["LINK"]) > 0
    assert "PIP" not in output["actions"]


@pytest.mark.integration
def test_update_env_only_pip_json_output(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Update an environment by adding only a pip package
    Check the json output
    """
    create_env(ENVIRONMENT_PYTHON_PIP_CLICK)
    conda_cli("env", "create", "--name", "envjson-4", "--json", "--yes")
    create_env(ENVIRONMENT_PYTHON_PIP_CLICK_ATTRS)
    stdout, _, _ = conda_cli(
        "env", "update", "--name", "envjson-4", "--quiet", "--json"
    )
    output = json.loads(stdout)
    assert output["success"] is True
    # No conda actions (FETCH/LINK), only pip
    assert list(output["actions"].keys()) == ["PIP"]
    # Only attrs installed
    assert len(output["actions"]["PIP"]) == 1
    assert output["actions"]["PIP"][0].startswith("attrs")


@pytest.mark.integration
def test_update_env_no_action_json_output(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Update an already up-to-date environment
    Check the json output
    """
    create_env(ENVIRONMENT_PYTHON_PIP_CLICK)
    conda_cli("env", "create", "--name", "envjson-5", "--json", "--yes")
    stdout, _, _ = conda_cli(
        "env", "update", "--name", "envjson-5", "--quiet", "--json"
    )
    output = json.loads(stdout)
    assert output["message"] == "All requested packages already installed."


@pytest.mark.integration
def test_remove_dry_run(env_name_1: None, conda_cli: CondaCLIFixture):
    # Test for GH-10231
    create_env(ENVIRONMENT_1)
    conda_cli("env", "create")
    conda_cli("env", "remove", "--name", "env-1", "--dry-run")
    assert env_is_created("env-1")


@pytest.mark.integration
def test_set_unset_env_vars(env_name_1: None, conda_cli: CondaCLIFixture):
    create_env(ENVIRONMENT_1)
    conda_cli("env", "create")
    env_name = "env-1"
    conda_cli(
        *("env", "config", "vars", "set"),
        *("--name", env_name),
        "DUDE=woah",
        "SWEET=yaaa",
        "API_KEY=AaBbCcDd===EeFf",
    )
    stdout, _, _ = conda_cli(
        *("env", "config", "vars", "list"),
        *("--name", env_name),
        "--json",
    )
    output_env_vars = json.loads(stdout)
    assert output_env_vars == {
        "DUDE": "woah",
        "SWEET": "yaaa",
        "API_KEY": "AaBbCcDd===EeFf",
    }

    conda_cli(
        *("env", "config", "vars", "unset"),
        *("--name", env_name),
        "DUDE",
        "SWEET",
        "API_KEY",
    )
    stdout, _, _ = conda_cli(
        *("env", "config", "vars", "list"),
        *("--name", env_name),
        "--json",
    )
    output_env_vars = json.loads(stdout)
    assert output_env_vars == {}


@pytest.mark.integration
def test_set_unset_env_vars_env_no_exist(env_name_1: None, conda_cli: CondaCLIFixture):
    create_env(ENVIRONMENT_1)
    conda_cli("env", "create")
    env_name = "env-11"
    with pytest.raises(EnvironmentLocationNotFound):
        conda_cli(
            *("env", "config", "vars", "set"),
            *("--name", env_name),
            "DUDE=woah",
            "SWEET=yaaa",
            "API_KEY=AaBbCcDd===EeFf",
        )


@pytest.mark.integration
def test_pip_error_is_propagated(env_name_1: None, conda_cli: CondaCLIFixture):
    """
    Creates an environment from an environment.yml file with conda and incorrect pip dependencies
    The output must clearly show pip error.
    Check the json output
    """
    create_env(ENVIRONMENT_PYTHON_PIP_NONEXISTING)
    with pytest.raises(CondaEnvException, match="Pip failed"):
        conda_cli("env", "create", "--name", TEST_ENV_NAME_PIP)


def env_is_created(env_name):
    """
        Assert an environment is created
    Args:
        env_name: the environment name
    Returns: True if created
             False otherwise
    """
    from os.path import basename

    for prefix in list_all_known_prefixes():
        name = ROOT_ENV_NAME if prefix == context.root_prefix else basename(prefix)
        if name == env_name:
            return True

    return False


@pytest.fixture
def env_name_2(conda_cli: CondaCLIFixture) -> None:
    # It *can* happen that this does not remove the env directory and then
    # the CREATE fails. Keep your eyes out! We could use rm_rf, but do we
    # know which conda install we're talking about? Now? Forever? I'd feel
    # safer adding an `rm -rf` if we had a `Commands.ENV_NAME_TO_PREFIX` to
    # tell us which folder to remove.
    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")

    yield

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")


@pytest.mark.integration
def test_env_export(
    env_name_2: None, conda_cli: CondaCLIFixture, path_factory: PathFactoryFixture
):
    """Test conda env export."""
    conda_cli("create", "--name", TEST_ENV_NAME_2, "flask", "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    stdout, _, _ = conda_cli("env", "export", "--name", TEST_ENV_NAME_2)

    env_yml = path_factory(suffix=".yml")
    env_yml.write_text(stdout)

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)
    conda_cli("env", "create", "--file", env_yml, "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    # regression test for #6220
    stdout, stderr, _ = conda_cli(
        "env", "export", "--name", TEST_ENV_NAME_2, "--no-builds"
    )
    assert not stderr
    env_description = yaml_safe_load(stdout)
    assert len(env_description["dependencies"])
    for spec_str in env_description["dependencies"]:
        assert spec_str.count("=") == 1

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)


@pytest.mark.integration
def test_env_export_with_variables(
    env_name_2: None, conda_cli: CondaCLIFixture, path_factory: PathFactoryFixture
):
    """Test conda env export."""
    conda_cli("create", "--name", TEST_ENV_NAME_2, "flask", "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    conda_cli(
        *("env", "config", "vars", "set"),
        *("--name", TEST_ENV_NAME_2),
        "DUDE=woah",
        "SWEET=yaaa",
    )

    stdout, _, _ = conda_cli("env", "export", "--name", TEST_ENV_NAME_2)

    env_yml = path_factory(suffix=".yml")
    env_yml.write_text(stdout)

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)
    conda_cli("env", "create", "--file", env_yml, "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    stdout, stderr, _ = conda_cli(
        "env", "export", "--name", TEST_ENV_NAME_2, "--no-builds"
    )
    assert not stderr
    env_description = yaml_safe_load(stdout)
    assert len(env_description["variables"])
    assert env_description["variables"].keys()

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)


@pytest.mark.integration
def test_env_export_json(env_name_2: None, conda_cli: CondaCLIFixture):
    """Test conda env export."""
    conda_cli("create", "--name", TEST_ENV_NAME_2, "flask", "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    stdout, _, _ = conda_cli("env", "export", "--name", TEST_ENV_NAME_2, "--json")

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)

    # regression test for #6220
    stdout, stderr, _ = conda_cli(
        "env", "export", "--name", TEST_ENV_NAME_2, "--no-builds", "--json"
    )
    assert not stderr

    env_description = json.loads(stdout)
    assert len(env_description["dependencies"])
    for spec_str in env_description["dependencies"]:
        assert spec_str.count("=") == 1

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)


@pytest.mark.integration
def test_list(
    env_name_2: None, conda_cli: CondaCLIFixture, path_factory: PathFactoryFixture
):
    """Test conda list -e and conda create from txt."""
    conda_cli("create", "--name", TEST_ENV_NAME_2, "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    stdout, _, _ = conda_cli("list", "--name", TEST_ENV_NAME_2, "--export")

    env_txt = path_factory(suffix=".txt")
    env_txt.write_text(stdout)

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)
    conda_cli("create", "--name", TEST_ENV_NAME_2, "--file", env_txt, "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    stdout2, _, _ = conda_cli("list", "--name", TEST_ENV_NAME_2, "--export")
    assert stdout == stdout2


@pytest.mark.integration
def test_export_multi_channel(
    env_name_2: None, conda_cli: CondaCLIFixture, path_factory: PathFactoryFixture
):
    """Test conda env export."""
    from conda.core.prefix_data import PrefixData

    PrefixData._cache_.clear()
    conda_cli("create", "--name", TEST_ENV_NAME_2, "python", "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    # install something from other channel not in config file
    conda_cli(
        "install",
        *("--name", TEST_ENV_NAME_2),
        *("--channel", "conda-test"),
        "test_timestamp_sort",
        "--yes",
    )
    stdout, _, _ = conda_cli("env", "export", "--name", TEST_ENV_NAME_2)
    assert "conda-test" in stdout

    stdout1, _, _ = conda_cli("list", "--name", TEST_ENV_NAME_2, "--explicit")

    env_yml = path_factory(suffix=".yml")
    env_yml.write_text(stdout)

    conda_cli("env", "remove", "--name", TEST_ENV_NAME_2, "--yes")
    assert not env_is_created(TEST_ENV_NAME_2)
    conda_cli("env", "create", "--file", env_yml, "--yes")
    assert env_is_created(TEST_ENV_NAME_2)

    # check explicit that we have same file
    stdout2, _, _ = conda_cli("list", "--name", TEST_ENV_NAME_2, "--explicit")
    assert stdout1 == stdout2


@pytest.mark.integration
def test_non_existent_file(env_name_2: None, conda_cli: CondaCLIFixture):
    with pytest.raises(EnvironmentFileNotFound):
        conda_cli("env", "create", "--file", "i_do_not_exist.yml", "--yes")


@pytest.mark.integration
def test_invalid_extensions(
    env_name_2: None,
    conda_cli: CondaCLIFixture,
    path_factory: PathFactoryFixture,
):
    env_yml = path_factory(suffix=".ymla")
    env_yml.touch()

    with pytest.raises(EnvironmentFileExtensionNotValid):
        conda_cli("env", "create", "--file", env_yml, "--yes")



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

from . import archspec, cuda, freebsd, linux, osx, windows

#: The list of virtual package plugins for easier registration with pluggy
plugins = [archspec, cuda, freebsd, linux, osx, windows]
